#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.ml - Texte et machine learning
# 
# Revue de méthodes de [word embedding](https://en.wikipedia.org/wiki/Word_embedding) statistiques (~ [NLP](https://en.wikipedia.org/wiki/Natural_language_processing)) ou comment transformer une information textuelle en vecteurs dans un espace vectoriel (*features*) ? Deux exercices sont ajoutés à la fin.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Données
# 
# Nous allons travailler sur des données twitter collectées avec le mot-clé macron : [tweets_macron_sijetaispresident_201609.zip](https://github.com/sdpython/ensae_teaching_cs/raw/master/src/ensae_teaching_cs/data/data_web/tweets_macron_sijetaispresident_201609.zip).

# In[2]:


from ensae_teaching_cs.data import twitter_zip
df = twitter_zip(as_df=True)
df.head(n=2).T


# In[3]:


df.shape


# 5000 tweets n'est pas assez pour tirer des conclusions mais cela donne une idée. On supprime les valeurs manquantes.

# In[4]:


data = df[["retweet_count", "text"]].dropna()
data.shape


# ## Construire une pondération
# 
# Le texte est toujours délicat à traiter. Il n'est pas toujours évident de sortir d'une information binaire : un mot est-il présent ou pas. Les mots n'ont aucun sens numérique. Une liste de tweets n'a pas beaucoup de sens à part les trier par une autre colonne : les retweet par exemple.

# In[5]:


data.sort_values("retweet_count", ascending=False).head()


# Sans cette colonne qui mesure la popularité, il faut trouver un moyen d'extraire de l'information. On découpe alors en mots et on constuire un modèle de langage : les [n-grammes](https://fr.wikipedia.org/wiki/N-gramme). Si un tweet est constitué de la séquence de mots $(m_1, m_2, ..., m_k)$. On définit sa probabilité comme :
# 
# $$P(tweet) = P(w_1, w_2) P(w_3 | w_2, w_1) P(w_4 | w_3, w_2) ... P(w_k | w_{k-1}, w_{k-2})$$
# 
# Dans ce cas, $n=3$ car on suppose que la probabilité d'apparition d'un mot ne dépend que des deux précédents. On estime chaque n-grammes comme suit :
# 
# $$P(c | a, b) = \frac{ \# (a, b, c)}{ \# (a, b)}$$
# 
# C'est le nombre de fois où on observe la séquence $(a,b,c)$ divisé par le nombre de fois où on observe la séquence $(a,b)$.

# ### Tokenisation
# 
# Découper en mots paraît simple ``tweet.split()`` et puis il y a toujours des surprises avec le texte, la prise en compte des tirets, les majuscules, les espaces en trop. On utilse un *tokenizer* dédié : [TweetTokenizer](http://www.nltk.org/api/nltk.tokenize.html#nltk.tokenize.casual.TweetTokenizer) ou un tokenizer qui prend en compte le langage.

# In[6]:


from nltk.tokenize import TweetTokenizer
tknzr = TweetTokenizer(preserve_case=False)
tokens = tknzr.tokenize(data.loc[0, "text"])
tokens


# ### n-grammes
# 
# * [N-Gram-Based Text Categorization: Categorizing Text With Python](http://blog.alejandronolla.com/2013/05/20/n-gram-based-text-categorization-categorizing-text-with-python/)

# In[7]:


from nltk.util import ngrams
generated_ngrams = ngrams(tokens, 4, pad_left=True, pad_right=True)
list(generated_ngrams)


# ### Exercice 1 : calculer des n-grammes sur les tweets

# In[8]:





# ### Nettoyage
# 
# Tous les modèles sont plus stables sans les stop-words, c'est-à-dire tous les mots présents dans n'importe quel documents et qui n'apporte pas de sens (à, de, le, la, ...). Souvent, on enlève les accents, la ponctuation... Moins de variabilité signifie des statistiques plus fiable.

# ### Exercice 2 : nettoyer les tweets
# 
# Voir [stem](http://www.nltk.org/api/nltk.stem.html#module-nltk.stem).

# ## Structure de graphe
# 
# On cherche cette fois-ci à construire des coordonnées pour chaque tweet.

# ### matrice d'adjacence

# Une option courante est de découper chaque expression en mots puis de créer une matrice *expression x mot* ou chaque case indique la présence d'un mot dans une expression.

# In[9]:


from sklearn.feature_extraction.text import CountVectorizer
count_vect = CountVectorizer()
counts = count_vect.fit_transform(data["text"])
counts.shape


# On aboutit à une matrice sparse ou chaque expression est représentée à une vecteur ou chaque 1 représente l'appartenance d'un mot à l'ensemble.

# In[10]:


type(counts)


# In[11]:


counts[:5,:5].toarray()


# In[12]:


data.loc[0,"text"]


# In[13]:


counts[0,:].sum()


# ### td-idf

# Ce genre de technique produit des matrices de très grande dimension qu'il faut réduire. On peut enlever les mots rares ou les mots très fréquents. [td-idf](https://fr.wikipedia.org/wiki/TF-IDF) est une technique qui vient des moteurs de recherche. Elle construit le même type de matrice (même dimension) mais associe à chaque couple (document - mot) un poids qui dépend de la fréquence d'un mot globalement et du nombre de documents contenant ce mot.
# 
# $$idf(t) = \log \frac{\# D}{\#\{d \; | \; t \in d \}}$$
# 
# Où :
# 
# * $\#D$ est le nombre de tweets
# * $\#\{d \; | \; t \in d \}$ est le nombre de tweets contenant le mot $t$

# $f(t,d)$ est le nombre d'occurences d'un mot $t$ dans un document $d$.
# 
# $$tf(t,d) = \frac{1}{2} + \frac{1}{2} \frac{f(t,d)}{\max_{t' \in d} f(t',d)}$$

# On construit le nombre $tfidf(t,f)$
# 
# $$tdidf(t,d) = tf(t,d) idf(t)$$
# 
# Le terme $idf(t)$ favorise les mots présent dans peu de documents, le terme $tf(t,f)$ favorise les termes répétés un grand nombre de fois dans le même document. On applique à la matrice précédente.

# In[14]:


from sklearn.feature_extraction.text import TfidfTransformer
tfidf = TfidfTransformer()
res = tfidf.fit_transform(counts)
res.shape


# In[15]:


res[0,:].sum()


# ### Exercice 3 : tf-idf sans mot-clés
# 
# La matrice ainsi créée est de grande dimension. Il faut trouver un moyen de la réduire avec [TfidfVectorizer](http://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html).

# In[16]:





# ### word2vec
# 
# * [word2vec From theory to practice](http://hen-drik.de/pub/Heuer%20-%20word2vec%20-%20From%20theory%20to%20practice.pdf)
# * [Efficient Estimation of Word Representations in Vector Space](https://arxiv.org/abs/1301.3781)
# * [word2vec](https://radimrehurek.com/gensim/models/word2vec.html)
# 
# Cet algorithme part d'une répresentation des mots sous forme de vecteur en un espace de dimension N = le nombre de mots distinct. Un mot est représenté par $(0,0, ..., 0, 1, 0, ..., 0)$. L'astuce consiste à réduire le nombre de dimensions en compressant avec une ACP, un réseau de neurones non linéaires.

# In[17]:


sentences = [tknzr.tokenize(_) for _ in data["text"]]
sentences[0]


# In[18]:


import gensim, logging
logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)

model = gensim.models.Word2Vec(sentences, min_count=1)


# In[19]:


model.wv.similar_by_word("fin")


# In[20]:


model.wv["fin"].shape


# In[21]:


model.wv["fin"]


# ## Tagging
# 
# L'objectif est de tagger les mots comme déterminer si un mot est un verbe, un adjectif ...

# ### grammar
# 
# Voir [html.grammar](http://www.nltk.org/api/nltk.html#module-nltk.grammar).

# ### CRF
# 
# Voir [CRF](http://www.nltk.org/api/nltk.tag.html)

# ### HMM
# 
# Voir [HMM](http://www.nltk.org/api/nltk.tag.html#module-nltk.tag.hmm).

# ## Clustering
# 
# Une fois qu'on a des coordonnées, on peut faire plein de choses.

# ### LDA
# 
# * [Latent Dirichlet Application](http://www.jmlr.org/papers/volume3/blei03a/blei03a.pdf)
# * [LatentDirichletAllocation](http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.LatentDirichletAllocation.html)

# In[22]:


from sklearn.feature_extraction.text import TfidfVectorizer

tfidf_vectorizer = TfidfVectorizer(max_df=0.95, min_df=2,
                                   max_features=1000)
tfidf = tfidf_vectorizer.fit_transform(data["text"])


# In[23]:


tfidf.shape


# In[24]:


from sklearn.decomposition import NMF, LatentDirichletAllocation
lda = LatentDirichletAllocation(n_components=10, max_iter=5,
                                learning_method='online',
                                learning_offset=50.,
                                random_state=0)


# In[25]:


lda.fit(tfidf)


# In[26]:


tf_feature_names = tfidf_vectorizer.get_feature_names()
tf_feature_names[100:103]


# In[27]:


def print_top_words(model, feature_names, n_top_words):
    for topic_idx, topic in enumerate(model.components_):
        print("Topic #%d:" % topic_idx)
        print(" ".join([feature_names[i]
                        for i in topic.argsort()[- n_top_words - 1:][::-1]]))
    print()


# In[28]:


print_top_words(lda, tf_feature_names, 10)


# In[29]:


tr = lda.transform(tfidf)
tr[:5]


# In[30]:


tr.shape


# In[31]:


import pyLDAvis
import pyLDAvis.sklearn
pyLDAvis.enable_notebook()


# In[32]:


pyLDAvis.sklearn.prepare(lda, tfidf, tfidf_vectorizer)


# In[33]:





# ### Exercice 4 : LDA
# 
# Recommencer en supprimant les stop-words pour avoir des résultats plus propres.

# In[34]:





# In[35]:





# In[36]: